@@ -3,7 +3,6 @@
 import asyncio
 import logging
 from functools import partial
-from pprint import pprint  # noqa: W0611
 
 from box import Box
 
@@ -34,7 +33,8 @@ class Component:
         self._paused = False
 
         self.settings = Box(self.configure(*args, **kwargs) or dict())
-        config_text = ' '.join(f'{k}={v}' for k, v in self.settings.items())
+        config_text = (' '.join(f'{k}={v}'
+                                for k, v in workflow.safe_settings(self.settings).items()))
         self.debug(f'Initialized {config_text}')
 
     def configure(self):
@@ -71,9 +71,7 @@ class Component:
         self.debug('Running')
 
     def _formatted(self, msg, *args):
-        # from random import randint
-        # prefix = f'{id(self)}:{randint(0, 10)}:{self.id} ' if self.id else ''
-        prefix = f'{self.id} [{id(self)}] ' if self.id else ''
+        prefix = f'{self.id} ' if self.id else ''
         msg = f'{prefix}{self.type} {msg}'
         return (msg,) + args
 
