@@ -3,7 +3,6 @@
 import asyncio
 import logging
 from collections import deque
-from pprint import pprint  # noqa: W0611
 from threading import Barrier, Thread
 
 import janus
@@ -11,7 +10,7 @@ from box import Box
 
 from .component import Component
 from .pipe import Manifold, Message
-from .utils import aiter
+from .utils import aiter, isdict, islist
 
 logger = logging.getLogger(__name__)
 
@@ -50,7 +49,7 @@ class Node(Component):
         coroutines = [super().start(*args)]
         self.layers = ([self.inbox] +
                        self.ifilters.ordered() +
-                       [self.spawn(self.processor)] +
+                       [self.spawn_processor()] +
                        self.ofilters.ordered() +
                        [self.outbox])
         for layer in self.layers:
@@ -80,12 +79,31 @@ class Node(Component):
 
             self.debug(f'Finished processing')
 
-    def spawn(self, processor):
-        if self.blocking or self.scale > 1:
-            processor = ThreadedProcessor(processor, scale=self.scale)
-        return processor
+    def spawn_processor(self):
+        processor = self.get_processor()
+        if self.blocking:
+            raise NotImplementedError(f'Node.blocking in {self}')
 
-    async def processor(self, messages):
+        if self.scale == 1:
+            return processor
+
+        elif self.scale > 1:
+            return MultiProcessor(self)
+
+    def get_processor(self):
+        if isinstance(self, WithRetry):
+            self._retryable = None
+            return self._retryable_processor
+
+        return self.processor
+
+    async def _retryable_processor(self, messages, id_=None):
+        if self._retryable is None:
+            self._retryable = RetryableMessages(self, messages)
+        async for channel, message in self.processor(self._retryable, id_):
+            yield channel, message
+
+    async def processor(self, messages, id_=None):
         """Process and yield new (channel, message)."""
         async for channel, message in messages:
             yield await self.process(channel, message)
@@ -96,12 +114,78 @@ class Node(Component):
     process.stub = True
 
     def drop(self, message):  # TODO: implement message accounting and leak detection
-        # self.outbox.drop(message)
         raise NotImplementedError(f'drop() in {self}')
 
-    # def flush(): pass
-    # def save(): pass
-    # def load(): pass
+    def merged_settings(self, message, key=None, msgmap=None):
+        """Return node settings, with values overridden from the message if present."""
+        settings = self.settings.to_dict()
+        message = message.to_dict()
+        if key:
+            settings = settings[key]
+            message = message[key]
+        if msgmap is None:
+            msgmap = list(settings)
+        if islist(msgmap):
+            msgmap = dict(zip(msgmap, msgmap))
+        elif not isdict(msgmap):
+            raise TypeError(f'Argument msgmap must be a mapping or a sequence, not {type(msgmap)}')
+
+        return Box(settings, **dict((arg, message[msgarg]) for arg, msgarg in msgmap.items()))
+
+
+class WithRetry:
+    pass
+
+
+class MultiProcessor:
+
+    def __init__(self, node):
+        self.node = node
+        self.processor = self.node.get_processor()
+
+    async def feeder(self, messages):
+        """Loop through all messages from the inbox and relay them to the input queue."""
+        async for channel, message in messages:
+            self.pending += 1
+            await self.iqueue.put((channel, message))
+        self.exhausted = True
+        if self.pending > 0:
+            await self.finished.wait()
+        for i in range(self.node.scale):
+            await self.iqueue.put(Message.EOT)
+
+    async def collector(self, id_):
+        """Launch an instance of the node processor, and relay its output to the output queue."""
+        async for channel, message in self.processor(aiter(self.iqueue.get, Message.EOT), id_=id_):
+            await self.oqueue.put((channel, message))
+        await self.oqueue.put((None, Message.EOT))
+
+    async def outfeed(self, messages):
+        """Launch all processor instances and return an iterator that feeds on the output queue."""
+        self.iqueue = asyncio.Queue(maxsize=self.node.scale)  # input queue
+        self.oqueue = asyncio.Queue(maxsize=1)                # output queue
+        self.finished = asyncio.Event()
+        self.exhausted = False
+        self.pending = 0  # number of messages currently handled by node processor
+        coroutines = [self.node.loop.create_task(self.feeder(messages))]
+        for i in range(self.node.scale):
+            coroutines.append(self.node.loop.create_task(self.collector(i)))
+        running = self.node.scale
+        while running or not self.oqueue.empty():
+            channel, message = await self.oqueue.get()
+            self.oqueue.task_done()
+            if message is Message.EOT:
+                running -= 1
+            else:
+                self.pending -= 1
+                if self.pending == 0 and self.exhausted:
+                    self.finished.set()
+                yield channel, message
+
+        for coroutine in coroutines:
+            await coroutine
+
+    __call__ = outfeed
 
 
 class Inbox(Manifold):
@@ -116,7 +200,6 @@ class Inbox(Manifold):
                 return
 
             try:
-                self.debug(f'* waiting for message on channel {channel}')
                 message = await pipe.receive()
             except Exception:
                 self.exception(f'Error while receiving on channel {channel}, {pipe}')
@@ -142,60 +225,10 @@ class Inbox(Manifold):
                 else:
                     yield channel, pipe
 
-    async def Xreceiver(self):
-        self.debug('Receiving')
-        self.debug(f'Setting up channels {list(self.channels)}')
-        # __import__('pudb').set_trace()
-        # import sys; sys.exit()
-        feeder = deque()
-        events = Box(received=asyncio.Event())
-        # receivers = dict()
-        for channel, pipe in self._active_channels():
-            self.debug(f'Setting up channel {channel}')
-            events[channel] = asyncio.Event()
-            events[channel].set()
-            # receivers[channel] = asyncio.ensure_future(
-            #     self.channel_receiver(channel, feeder, events))
-            asyncio.ensure_future(self.channel_receiver(channel, feeder, events))
-
-        self.debug('Waiting for feeder')
-        while self.running and (len(feeder) or await events.received.wait()):
-            for channel in self.channels:
-                events[channel].clear()
-            events.received.clear()
-            channel, message = feeder.popleft()
-
-            self.debug(f'Got message from {channel}: {message}')
-            if message == Message.EOT:
-                self.debug('Exhausted channel: %s' % channel)
-                pipe.stop()
-                self.debug(f'* setting event for channel {channel}')
-                events[channel].set()
-                # receivers[channel].cancel()
-                try:
-                    await pipe.send(Message.EOT)
-                except NotImplementedError:
-                    pass
-                self.debug('-----------------------------')
-                if not list(self._active_channels()):
-                    self.debug('Stopping: no more active channels')
-                    self.stop()
-                    # TODO: cancel channel receiver coroutines
-                    break
-            else:
-                message.checkin(self)
-                yield channel, message
-
-            for channel in self.channels:
-                events[channel].set()
-
-        self.debug('Finished receiving')
-
     async def channel_receiver(self, channel, feeder, events):
         self.debug(f'Receiver for {channel} is ready')
         pipe = self.channels[channel]
         while pipe.running and self.running and await events[channel].wait():
-            self.debug(f'* receiving from channel {channel}')
             events[channel].clear()
             try:
                 feeder.append((channel, await pipe.receive()))
@@ -203,30 +236,64 @@ class Inbox(Manifold):
                 self.exception(f'Error while receiving on channel {channel}, {pipe}')
                 raise
 
-            self.debug(f'* received from channel {channel} (next event: {events[channel].is_set()})')
             events.received.set()
         self.debug(f'Receiver for {channel} is exiting')
 
-    def X_active_channels(self):
-        channels = self.channels.copy()
-        while channels:
-            for channel, pipe in channels.copy().items():
-                del channels[channel]
-                if not pipe.running:
-                    self.debug(f'skipping stopped channel {channel}')
-                else:
-                    yield channel, pipe
-        if channels:
-            pass  # TODO: channel clean-ups
 
-    async def _retry(self, channel, message, wait=0):
-        # TODO: keep track of retried messages (for checkpointing a node)
-        await asyncio.sleep(wait)                   # maybe the message shouldn't be pushed
-        await self.channels[channel].send(message)  # to the bottom of a potentially long queue
+class RetryableMessages:
+    """Interfaces a queue between the inbox and the processor where we can insert back retries."""
 
-    def retry(self, channel, message, wait=0):
-        self.debug(f'Retrying in {wait}s on {channel}: {message}')
-        asyncio.run_coroutine_threadsafe(self._retry(channel, message, wait), loop=self.loop)
+    def __init__(self, node, messages):
+        self.node = node
+        self.messages = messages
+        self._forwarder = None
+
+    async def __aiter__(self):
+        if self._forwarder is None:  # singleton forwarder, for when we're using MultiProcessor
+            self._forwarder = self.forwarder()
+        return self._forwarder
+
+    async def feeder(self):
+        """Feed the inbox to the queue."""
+        async for channel, self.newmsg in self.messages:
+            self.forwarded.clear()
+            await self.queue.put((channel, self.newmsg))
+            await self.forwarded.wait()
+        self.exhausted = True
+        if self.pending > 0:
+            await self.finished.wait()
+        await self.queue.put(Message.EOT)
+
+    async def forwarder(self):
+        """Turn the queue into an iterator that the processor can consume."""
+        self.queue = asyncio.Queue()
+        self.forwarded = asyncio.Event()
+        self.finished = asyncio.Event()
+        self.exhausted = False
+        self.newmsg = None
+        self.pending = 0  # number of messages currently handled by node processor
+        feeder = self.node.loop.create_task(self.feeder())
+        async for channel, message in aiter(self.queue.get, Message.EOT):
+            self.queue.task_done()
+            if message is self.newmsg:
+                self.forwarded.set()
+            self.pending += 1
+            yield channel, message
+
+        self.queue.task_done()
+        await feeder
+
+    def processing_done(self):
+        """Signal that a message is not in the processor anymore."""
+        self.pending -= 1
+        if self.exhausted:
+            self.finished.set()
+
+    async def retry(self, channel, message, wait=0.001, id_=None):
+        """Insert a message back into the queue for retrying it."""
+        await asyncio.sleep(wait)
+        await self.node.loop.create_task(self.queue.put((channel, message)))
+        self.processing_done()
 
 
 class Outbox(Manifold):
@@ -243,11 +310,9 @@ class Outbox(Manifold):
                 self.debug(f'Sending to {channel}: {message}')
                 if channel in self.channels:
                     await self.channels[channel].send(message)
-                    self.debug(f'Sent to    {channel}: {message} [{self.channels[channel].id}]')
                     message.checkout(self)
                     yield channel, message
 
-                    self.debug(f'Yielded    {channel}: {message}')
                 else:
                     raise KeyError(f'Channel "{channel}" does not exist in {self}')
 
@@ -317,34 +382,25 @@ class ThreadedProcessor:
             if i:  # skip the one already consumed by aiter above
                 await self.oqueue.async_q.get()
             self.oqueue.async_q.task_done()
-            # print(f'thread {i} EOT received back')
 
-        while self.threads:  # FIXME: join threads instead, and move or retire del self.threads[]
-            asyncio.sleep(0.1)  #     while at it, catch exceptions and clean up queue with task_done  # noqa: E501
-        # print(f'all threads stopped')
+        while self.threads:
+            asyncio.sleep(0.1)
 
     async def thread_feeder(self, messages):
         async for (channel, message) in messages:
-            # TODO: recreate dead threads, report thread failures
             await self.iqueue.async_q.put((channel, message))
         for i in range(self.scale):
             await self.iqueue.async_q.put((None, Message.EOT))
-            # print(f'thread {i} EOT sent')
 
     def thread_consumer(self, threadnum):  # runs in thread
-        # print(f'thread {threadnum} started')
         queue_iter = iter(self.iqueue.sync_q.get, (None, Message.EOT))
         for channel, message in self.processor(queue_iter):
             self.iqueue.sync_q.task_done()
             self.oqueue.sync_q.put((channel, message))
-        # print(f'thread {threadnum} EOT received')
         self.iqueue.sync_q.task_done()  # for EOT which does not make it past queue iterator
-        # print(f'thread {threadnum} task done')
         self.barrier.wait()
-        self.oqueue.sync_q.put((None, Message.EOT))  # TODO: use threading.Event to send just one
-        # print(f'thread {threadnum} EOT sent back')
+        self.oqueue.sync_q.put((None, Message.EOT))
         del self.threads[threadnum]
-        # print(f'thread {threadnum} stopped')
 
 
 class PriorityRegistry(dict):
