@@ -6,7 +6,6 @@ import sys
 from importlib import import_module
 from inspect import isclass
 from pathlib import Path
-from pprint import pprint  # noqa: W0611
 from string import Template
 
 import strictyaml
@@ -42,8 +41,7 @@ class Workflow:
                 raise ConfigurationException(
                     f'Missing environment variable "{e.args[0]}" (needed by {source})')
 
-        self.definition = self.expand(Box(strictyaml.load(source).data,
-                                          box_dots=True))
+        self.definition = self.expand(Box(strictyaml.load(source).data, box_dots=True))
 
     __MISSING__ = object()
 
@@ -59,9 +57,6 @@ class Workflow:
         elif isdict(element):
             useref = element.pop('<use>', None)
             if useref:
-                # if useref in seen:
-                #     raise ConfigurationException(f'Circular reference detected: {useref}')
-
                 seen.add(useref)
                 try:
                     useref = type(element)(settings[useref].to_dict())
@@ -87,23 +82,17 @@ class Workflow:
     def configure(self):
         """Locate and configure components."""
         _l.debug('Configuring workflow')
-        # pprint(self.app)
         for wname, workflow in self.app.workflows.items():
-            # print(f'workflow: {wname}')
             for nname, node in workflow.items():
                 node.workflow = wname
                 node.key = f'{wname}.{nname}'
-                # print(f'  node: {node.key}')
                 self._configure_node(node)
                 for msgbox in ('inbox', 'outbox'):
                     self._configure_pipes(node, msgbox)
         for wname, workflow in self.app.workflows.items():
-            # print(f'workflow: {wname}')
             for node in workflow.values():
                 for msgbox in ('inbox', 'outbox'):
                     self._configure_connections(node, msgbox)
-        # pprint(self.app)
-        # import sys; sys.exit()
 
     def _configure_node(self, node):
         try:
@@ -114,8 +103,6 @@ class Workflow:
             raise ConfigurationException(errmsg)
 
     def _configure_pipes(self, node, msgbox):
-        # _l.debug(f'  * {msgbox}: {node.key}')
-
         # Get default channels from class, if any.
         defaults = getattr(node.component, msgbox.upper(), dict())
         if defaults:
@@ -131,7 +118,7 @@ class Workflow:
         if isstr(defined_channels) and defined_channels.startswith('<'):
             return  # reference will be later resolved in _configure_connections()
 
-        # Expand shortened configuration structure, if relevent.
+        # Expand shortened configuration structure, if relevant.
         channels = Box(defaults)
         if ischannel(defined_channels):
             channels.update(default=defined_channels)
@@ -139,7 +126,6 @@ class Workflow:
             channels.update(**defined_channels)
         if not channels:
             channels.update(default=None)
-        # print(' mb 1', node.key, msgbox, channels)
 
         # Resolve pipe specs to classes.
         for name, spec in channels.items():
@@ -147,8 +133,6 @@ class Workflow:
             if not spec:
                 spec = 'pipekit.pipe:QueuePipe'
             if isstr(spec):
-                # if spec == '<void>':
-                #     spec = 'pipekit.pipe:NullPipe'
                 if spec.startswith('<'):
                     channel.node = spec.strip('< ')
                 else:
@@ -163,28 +147,19 @@ class Workflow:
                 channel.component = resolve(channel.component)
             channels[name] = channel
 
-        # print(' mb 2', channels)
         node[msgbox] = channels
-        # _l.debug(f'  ** {msgbox}: {node.key}')
 
     def _configure_connections(self, node, msgbox):
-        # _l.debug(f'  * {msgbox}: {node.key}')
-
         # Import all channels from peer node, if configured in.
         defined_channels = node.get(msgbox, {})
         if isstr(defined_channels) and defined_channels.startswith('<'):
-                # not defined_channels.endswith('>'):
-            # __import__('pudb').set_trace()
             spec = defined_channels.strip('< ')
             peer_node, _ = self.peer_node(spec, node)
             peer_msgbox = 'outbox' if msgbox == 'inbox' else 'inbox'
-            # _l.debug(f'  ** new PipeRefs: {node.key} < {peer_node.key}.{peer_msgbox}')
             channels = Box((c, PipeRef(peer_node, peer_msgbox, c))
                            for c in getattr(peer_node, peer_msgbox))
 
-            # print(' mb 2', channels)
             node[msgbox] = channels
-        # _l.debug(f'  ** {msgbox}: {node.key}')
 
     def build(self):
         """Instantiate and wire up nodes and pipes."""
@@ -200,8 +175,6 @@ class Workflow:
             except Exception:
                 _l.exception(f'Error while instantiating node {node.key}')
                 raise
-        # pprint(self.app)
-        # import sys; sys.exit()
 
     RESERVED_SETTINGS = set(['id', 'process', 'scale', 'inbox', 'ifilters', 'ofilters', 'outbox'])
 
@@ -210,7 +183,6 @@ class Workflow:
         if node.key not in self._node_backlog:
             return
 
-        # _l.debug(f'  node: {node.key}')
         settings = node.get('settings', {})
         reserved_settings = self.RESERVED_SETTINGS.intersection(set(settings.keys()))
         if reserved_settings:
@@ -226,23 +198,17 @@ class Workflow:
 
         inbox = dict()
         for channel, pipe in node.inbox.items():
-            # _l.debug(f'  inbox.{channel}: {pipe}')
             if hasattr(pipe, 'instance'):
-                # _l.debug(f'  * instance exists: {pipe.instance}')
                 inbox[channel] = pipe.instance
                 raise Exception('node instance exists')
 
             if isinstance(pipe, PipeRef):
-                # _l.debug(f'  * connection {pipe}')
                 node.inbox[channel] = pipe = Box(component=pipe, instance=pipe.resolve())
-                # _l.debug(f'  * connected {pipe}')
-            else:  # if 'component' in pipe:
-                # _l.debug(f'  * component type: {type(pipe.component)}')
+            else:
                 pipe.instance = self.make_component(
                     pipe.component, id=f'{node.key}.input.{channel}', **pipe.get('settings', {}))
             inbox[channel] = pipe.instance
 
-        # TODO: define and instantiate filters
         ifilters = None
         ofilters = None
         node_args = dict(
@@ -259,7 +225,6 @@ class Workflow:
                 f'got {type(node.component)} instead')
 
         node.instance = self.make_component(node_class, **node_args)
-        # print(f'    instance: {node.instance}')
         del self._node_backlog[node.key]
 
     def peer_node(self, spec, dependent):
@@ -278,10 +243,26 @@ class Workflow:
         return node, channel
 
     def make_component(self, class_, *args, **kwargs):
-        # _l.debug(f'Making new component {class_} id={kwargs["id"]}')
         return class_(self, *args, **kwargs)
 
+    _SECRETS = set('account password secret'.split())
+
+    def safe_settings(self, settings=__MISSING__):
+        """Return modified settings where secrets have been hidden."""
+        if settings is self.__MISSING__:
+            settings = self.app
+        settings = settings.copy()
+        for key, value in settings.items():
+            if isdict(value):
+                settings[key] = self.safe_settings(value)
+            elif islist(value):
+                settings[key] = [self.safe_settings(i) for i in value]
+            elif key in self._SECRETS:
+                settings[key] = '<secret>'
+        return settings
+
     def run(self):
+        """Create engine and run workflow."""
         self.engine = ETLEngine(self)
         self.engine.run()
 
@@ -290,13 +271,6 @@ def resolve(spec):
     """Parse component textual spec and import and return corresponding object."""
     try:
         module, attr, *_ = spec.rsplit(':', 1) + [None]
-        # nameparts = module.split('.')
-        # if nameparts[0] == 'pipekit':
-        #     module = '.'.join(nameparts[1:])
-        #     parent = 'pipekit'
-        # else:
-        #     parent = 'components'
-        # module = import_module(f'{parent}.{module}')
         module = import_module(module)
         return getattr(module, attr, None)
 
